/*
 *   General double-linked list library
 *
 *   (c) Gal "Kilobug" Le Mignot <kilobug@hurdfr.org>, 2002
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 *   $Id: list.c,v 1.4 2002/10/17 17:50:16 kilobug Exp $
 *
 */



#include "list.h"



/* 
   Creation, destruction
*/

/* Creates a new list with the following data */
list_t 
list_create (void *data)
{
  list_t res = malloc (sizeof (*res));

  if (res) 
    {
      res->prev = res->next = NULL;
      res->data = data;
    }
  
  return res;
}

/* Free the list */
void 
list_free (list_t list)
{
  if (!list)
    return;
  list_free (list->next);
  free (list);
}



/* 
   Misc functions
*/

/* Count the number of elements in the list */
size_t 
list_length (list_t list)
{
  size_t size = 0;
  
  for (list_t l = list; l != NULL; l = l->next)
    size++;
  return size;
}

/* Reverse the list */
list_t 
list_reverse (list_t list)
{
  list_t head = list;
  list_t queue = NULL;
  
  while (head != NULL)
    {
      list_t next = head->next;
      head->next = queue;
      if (queue)
	queue->prev = head;
      queue = head;      
      head = next;
    }
  if (queue)
    queue->prev = NULL;
  return queue;
}

/* Apply a function on all elements of the list. If func, ever returns
   non-zero, then iteration stops and list_iterate returns that value,
   otherwise it (eventually) returns 0. */
error_t 
list_iterate(list_t list, error_t (*func)(void *))
{
  for (list_t l = list; l != NULL; l = l->next)
    {
      error_t err = func (l->data);
      if (err)
	return err;
    }
  return 0;
}

/* Apply a function on all elements of the list - in this one the function 
   receives an extra, user-supplied void *. */
error_t 
list_iterate_with_data(list_t list, error_t (*func)(void *, void *), 
		       void *data)
{
  for (list_t l = list; l != NULL; l = l->next)
    {
      error_t err = func (l->data, data);
      if (err)
	return err;
    }
  return 0;
}



/* 
   Find functions
*/

/* Finds an element in the list, by comparing the data */
list_t 
list_find (list_t list, void *data)
{
  for (list_t l = list; l != NULL; l = l->next)
    {
      if (l->data == data)
	return l;
    }
  return NULL;
}


/* Finds and element in the list, using the given comparaison function  */
list_t 
list_find_custom (list_t list, void *data, list_compare_t compar)
{
  for (list_t l = list; l != NULL; l = l->next)
    {
      if (!compar(data, l->data))
	return l;
    }
  return NULL;
}

/* Locates item in the list list, returns -1 if not found */
int 
list_position (list_t list, void *item)
{
  int i = 0;
  
  for (list_t l = list; l != NULL; l = l->next)
    {
      if (l->data == item)
	return i;
      i++;
    }
  return -1;
}

/* Locates item in the list list, returns -1 if not found */
int 
list_position_custom (list_t list, void *data, list_compare_t compar)
{
  int i = 0;
  
  for (list_t l = list; l != NULL; l = l->next)
    {
      if (!compar(data, l->data))
	return i;
      i++;
    }
  return -1;
}


/* Returns the element number index in the list */
list_t 
list_index (list_t list, int index)
{
  list_t l = list;
  
  for (int i = 0; (i < index) && (l != NULL); i++)
    l = l->next;
  return l;
}



/* 
   Inserting
*/

/* Prepends an element to the beginning of list */
list_t 
list_prepend (list_t source, void *data)
{
  list_t res = list_create (data);

  if (!res)
    return NULL;
  
  if (source)
    {      
      res->next = source;
      source->prev = res;
    }
  return res;
}

/* Appends an element to the end of list - warning, this is O(n) */
list_t 
list_append (list_t source, void *data)
{
  list_t res = list_create (data);

  if (!res)
    return NULL;

  if (source)
    {
      list_t l = source;
      while (l->next != NULL)
	l = l->next;
      l->next = res;
      res->prev = l;
      return source;
    }

  return res;
}

/* Inserts an element to the list just after the given item */
error_t 
list_insert_after (list_t where, void *data)
{
  list_t l = NULL;
  
  if (where == NULL)
    return EINVAL;
  l = list_create (data);
  if (l == NULL)
    return errno;
  
  /* Connect the end of the list */
  l->next = where->next;
  if (l->next)
    l->next->prev = l;
  /* Connect the begining of the list */
  l->prev = where;
  l->prev->next = l;

  return 0;
}

/* Inserts an element to the list just before the given item */
error_t 
list_insert_before (list_t where, void *data)
{
  list_t l = NULL;
  
  if (where == NULL)
    return EINVAL;
  l = list_create (data);
  if (l == NULL)
    return errno;
  
  /* Connect the end of the list */
  l->next = where;
  l->next->prev = l;
  /* Connect the begining of the list */
  l->prev = where->prev;
  if (l->prev)
    l->prev->next = l;

  return 0;
}

/* Concatenate two lists */
list_t 
list_concat (list_t l1, list_t l2)
{
  list_t l;

  if (!l1)
    return l2;
  if (!l2)
    return l1;
  
  for (l = l1; l->next != NULL; l = l->next);
  l->next = l2;
  l2->prev = l;
  return l1;
}



/* 
   Deleting
*/

/* Removes an item from the list, looking by data - this is O(n) */
list_t 
list_remove_data (list_t list, void *data)
{
  list_t l;
  
  if (!list)
    return NULL;
  
  /* Special case: deletion of the first element */
  if (list->data == data)
    {
      l = list->next;
      free (list);
      return l;
    }

  for (l = list->next; l != NULL; l = l->next)
    {
      if (l->data == data)
	{
	  list_remove (l);
	  break;
	}
    }
  return list;
}

/* Unlinks the given element from its list, and makes it a stand-alone list */
error_t 
list_remove_link (list_t link)
{
  if (!link)
    return EINVAL;
  if (link->prev)
    link->prev->next = link->next;
  if (link->next)
    link->next->prev = link->prev;
  link->prev = link->next = NULL;

  return 0;
}

/* Unlinks the given element from its list, and destroys it */
error_t 
list_remove (list_t link)
{
  error_t err = list_remove_link (link);
  if (err)
    return err;
  free (link);
  
  return 0;
}



/* 
   Sorting
*/

static list_t
list_sort_with_head (list_t source, list_t *head, list_compare_t compar)
{
  void *cutpoint = NULL;
  list_t below = NULL, above = NULL;
  list_t below_head;
  list_t l = source;
  
  if ((!source) || (!source->next))
    {
      *head = source;
      return source;
    }
  
  /* Create the sublists */
  cutpoint = source->data;
  while (l != NULL)
    {
      list_t next = l->next;
      l->prev = NULL;
      if (compar (cutpoint, l->data) <= 0)
	{
	  l->next = below;
	  below->prev = l;
	  below = l;
	}
      else
	{
	  l->next = above;
	  above->prev = l;
	  above = l;
	}      
      l = next;
    }

  /* Put the termination */
  if (above)
    above->prev = NULL;
  if (below)
    below->prev = NULL;
  
  /* Sort the sublists */
  below = list_sort_with_head (below, &below_head, compar);
  above = list_sort_with_head (above, head, compar);

  /* Do the merge */
  if (!below_head)
    return above;
  below_head->next = above;
  if (above)
    above->prev = below_head;
  return below;
}

/* Sort the given list */
list_t 
list_sort (list_t source, list_compare_t compar)
{
  list_t head;
  
  return list_sort_with_head (source, &head, compar);
}

/* Insert an element in a sorted list */
list_t 
list_insert_sorted (list_t list, void *data, list_compare_t compar)
{
  list_t l = NULL;

  if (!list)
    return list_create (data);
  
  for (l = list; l->next != NULL; l = l->next)
    if (compar (data, l->next->data) >= 0)
      break;
  list_insert_after (l, data);
  return list;
}

/* Merge two sorted lists and keep the order */
list_t 
list_merge_sorted (list_t l1, list_t l2, list_compare_t compar)
{
  list_t res, p = NULL;
  
  while (l1 && l2)
    {
      /* Find who to insert now */
      list_t *q;
      if (compar(l1->data, l2->data) > 0)
	q = &l2;
      else
	q = &l1;

      /* Rebuild links */
      (*q)->prev = p;
      if (p == NULL)
	{
	  p = *q;
	  res = p;
	}
      else
	{
	  p->next = *q;
	  p = *q;
	}

      (*q) = (*q)->next;
    }

  /* Finish the link */
  p->next = l1 ? : l2;
  if (p->next)
    p->next->prev = p;
  return res;
}

