/*
 *   General double-linked list library
 *
 *   (c) Gal "Kilobug" Le Mignot <kilobug@hurdfr.org>, 2002
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 *   $Id: list.h,v 1.5 2002/10/17 17:50:16 kilobug Exp $
 *
 */



#ifndef __LIST_H__
#define __LIST_H__

#include <errno.h>
#include <stdlib.h>



typedef struct _list *list_t;

struct _list
{
  /* Pointer to the next element */
  list_t next;
  /* Pointer to the previous element */
  list_t prev;
  /* The data */
  void *data;
};

typedef int (*list_compare_t)(const void *d1, const void *d2);



/* 
   Creation, destruction
*/

/* Creates a new list with the following data */
list_t list_create (void *data);

/* Free the list */
void list_free (list_t list);



/* 
   Misc functions
*/

/* Count the number of elements in the list */
size_t list_length (list_t list);

/* Reverse the list */
list_t list_reverse (list_t list);

/* Apply a function on all elements of the list. If func, ever returns
   non-zero, then iteration stops and list_iterate returns that value,
   otherwise it (eventually) returns 0. */
error_t list_iterate(list_t list, error_t (*func)(void *));

/* Apply a function on all elements of the list - in this one the function 
   receives an extra, user-supplied void *. */
error_t list_iterate_with_data(list_t list, error_t (*func)(void *, void *), 
			       void *data);




/* 
   Find functions
*/

/* Finds an element in the list, by comparing the data */
list_t list_find (list_t list, void *data);

/* Finds and element in the list, using the given comparaison function  */
list_t list_find_custom (list_t list, void *data, list_compare_t compar);

/* Locates item in the list list, returns -1 if not found */
int list_position (list_t list, void *item);

/* Locates item in the list list, returns -1 if not found */
int list_position_custom (list_t list, void *data, list_compare_t compar);

/* Returns the element number index in the list */
list_t list_index (list_t list, int index);



/* 
   Inserting
*/

/* Prepends an element to the beginning of list */
list_t list_prepend (list_t source, void *data);

/* Appends an element to the end of list - warning, this is O(n) */
list_t list_append (list_t source, void *data);

/* Inserts an element to the list just after the given item */
error_t list_insert_after (list_t where, void *data);

/* Inserts an element to the list just before the given item */
error_t list_insert_before (list_t where, void *data);

/* Concatenate two lists */
list_t list_concat (list_t l1, list_t l2);



/* 
   Deleting
*/

/* Removes an item from the list, looking by data - this is O(n) */
list_t list_remove_data (list_t list, void *data);

/* Unlinks the given element from its list, and makes it a stand-alone list */
error_t list_remove_link (list_t link);

/* Unlinks the given element from its list, and destroys it */
error_t list_remove (list_t link);



/* 
   Sorting
*/

/* Sort the given list */
list_t list_sort (list_t source, list_compare_t compar);

/* Insert an element in a sorted list */
list_t list_insert_sorted (list_t list, void *data, list_compare_t compar);

/* Merge two sorted lists and keep the order */
list_t list_merge_sorted (list_t l1, list_t l2, list_compare_t compar);

#endif
